;;; -*- asm -*-
;;;
;;; Copyright (c) 2004  Robert James Kaes <rjk@wormbytes.ca>
;;;   All rights reserved.
;;;
;;;
;;; This program is free software; you can redistribute it and/or modify it
;;;  under the terms of the GNU General Public License as published by the
;;; Free Software Foundation; either version 2, or (at your option) any
;;; later version.
;;;
;;; This program is distributed in the hope that it will be useful, but
;;; WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;;  General Public License for more details.
;;; 

	include "processor.inc"

        include "display.inc"
        include "division.inc"

	;; Set up the various configuration information
        __CONFIG        _CONFIG1H, _OSCS_OFF_1H & _HSPLL_OSC_1H
        __CONFIG        _CONFIG2L, _BOR_OFF_2L & _PWRT_ON_2L
        __CONFIG        _CONFIG2H, _WDT_OFF_2H
        __CONFIG        300004h, 0xff
        __CONFIG        300005h, 0xff
        __CONFIG        _CONFIG4L, _DEBUG_OFF_4L & _LVP_OFF_4L
        __CONFIG        300007h, 0xff
        __CONFIG        _CONFIG5L, 0xff
        __CONFIG        _CONFIG5H, 0xff
        __CONFIG        _CONFIG6L, 0xff
        __CONFIG        _CONFIG6H, 0xff
        __CONFIG        _CONFIG7L, 0xff
        __CONFIG        _CONFIG7H, 0xff


;;; --------------------------------------------------------------------
;;;
;;; Reset Vector (startup)
;;;
.reset_vector	code	0h
	goto	main

;;; --------------------------------------------------------------------
;;;
;;; Interrupt service routine
;;;
.hi_int_vector	code	8h
        call    timer.isr
	retfie	FAST		; restore the registers


;;; --------------------------------------------------------------------
;;;
;;; Timer code (and counter)
;;;
.timer_data     udata
        ;; keep track of the number of times through the loop
        ;; without an RPM pulse.
rpm_wait_count  res     2
rpm_dividend    res     2

.timer_code     code

;;; RPM is defined as RPM = 6000 / rpm_wait_count
;;; where rpm_wait_count is incremented every 100us
#define RPM_CONSTANT_DIVIDEND   6000
#define RPM_UPDATE_TIMER        100

;;; We can't display more than 9.9 on the 7-segment display; however,
;;; if 100 we'll display a special "OL" message
#define RPM_MAX_DISPLAY         100

timer.isr:
        global  timer.isr
        
	;; check to see if the timer2 is enabled
	btfsc	PIR1, TMR2IF	; if not set, skip down to the return
	btfss	PIE1, TMR2IE	; if set, skip the return
	return

	bcf	PIR1, TMR2IF	; clear interrupt

        movfw   TMR0L
        bnz     calculate_rpm

        ;; not a new pulse, so add to the wait state
update_rpm_count:
        banksel rpm_wait_count
        meminc  rpm_wait_count, 2
        return

        ;; calculate the RPM since we received the pulse
calculate_rpm:
        subwf   TMR0L, F        ; update the counter (likely clearing it)
        
        ;; load the constant into the dividend
        movlw   LOW(RPM_CONSTANT_DIVIDEND)
        movwf   rpm_dividend
        movlw   HIGH(RPM_CONSTANT_DIVIDEND)
        movwf   rpm_dividend + 1 ; MSB

        ;; do the division now.  the results will be in rpm_dividend
        div1515u        rpm_dividend, rpm_wait_count

        ;; copy the result to the global RPM value
        banksel rpm_dividend
        movlw   RPM_MAX_DISPLAY ; limit how high the display can go
        cpfsgt  rpm_dividend    ; skip if rpm_dividend > RPM_MAX_DISPLAY
        bra     update_rpm

        ;; W is still RPM_MAX_DISPLAY so, set the max in the current_rpm
        banksel current_rpm
        movwf   current_rpm
        bra     clear_counter

update_rpm:     
        movff   rpm_dividend, current_rpm

clear_counter:  
        ;; clear the wait counter now
        banksel rpm_wait_count
        clrf    rpm_wait_count
        clrf    rpm_wait_count + 1

        ;; finished
        return


;;; ------------------------------------------------------------------
;;;
;;; Turn on the timer with a 100us timer interrupt
;;;
;;; PR2 = (OSC_FREQ / 4M) * (t(us) / (pre * post))
;;;

;;; PR2 = (40M / 4M) * (100 / (4 * 10)) = 25
#define TIMER_100U      25
        
timer.start:
        global	timer.start

	clrf	TMR2		; clear the timer to begin with
	movlw	TIMER_100U	; generate an interrupt every 100us
	movwf	PR2
	bsf	PIE1, TMR2IE	; enable the timer interrupt
	;; 1:4 pre, 1:10 post, turn on
	movlw	b'01001101'
	movwf	T2CON

        bsf     INTCON, GIE     ; enable global interrupts
        bsf     INTCON, PEIE    ; enable peripheral ints
	return
        
        

;;; --------------------------------------------------------------------
;;; Data
;;;
.main_data      udata
current_rpm     res     1       ; the current RPM calculated
copy_rpm        res     1       ; a waiting copy of the rpm
        
count   res     1
temp    res     1
delay   res     2

;;; --------------------------------------------------------------------
;;;
;;; Start of the program
;;;
.main_code	code

porta_init:
	movlw	07h		; configure a/d as digital inputs
	movwf	ADCON1

	;; set up RA2 and RA3 as outputs since they'll be sinking
	;; the LEDs
	movlw	b'11110011'
	movwf	TRISA
        clrf    PORTA
	return

portb_init:
	;; portb are _all_ outputs
	clrf	PORTB
	clrf	TRISB		; start off
	return
	
portc_init:
        ;; portc are _all_ outputs
        clrf    PORTC
        clrf    TRISC
	return

;;; --------------------------------------------------------------------
;;;
;;; Initialize the TMR0 counter
;;; 
counter_init:
        ;; TMR0ON = 1 = start timer
        ;; T08BIT = 1 = 8 bit timer/counter
        ;; T0CS   = 1 = transition on T0CKI pin
        ;; T0SE   = 0 = increment on low to high transition
        ;; PSA    = 1 = timer0 prescaler is NOT assigned
        ;; T0PS2:T0PS0 = prescaler bits (not used)
        movlw   b'11101111'
        movwf   T0CON

        ;; clear the counter
        clrf    TMR0L
	return

main:
	;; Initialize a few registers to make sure they're in a known
	;; good state.
	clrf	STATUS		; init the status register
	clrf	INTCON		; disable interrupts

	;; initialize the ports
	call	porta_init
	call	portb_init
	call	portc_init

        ;; Start with a zeroed RPM
        banksel current_rpm
        clrf    current_rpm

        ;; initialize the counter
        call    counter_init

        ;; start the timer
        call    timer.start

        ;; Continously update the display
update_display_loop:
        banksel current_rpm
        movfw   current_rpm     ; get the current rpm

        movff   current_rpm, copy_rpm
        clrf    delay

timed_display_loop:     
        banksel copy_rpm
        movfw   copy_rpm
        call    display_rpm

        decfsz  delay, F
        bra     timed_display_loop
       
        bra     update_display_loop
        
	end
	